/*************************************************************
 *                 bbomberman                                *
 *  (c) 2004 by Bernhard Eder (bb-generation)                *
 *  http://bbgen.net/bomberman for updates                   *
 *  Webmaster@bbgen.net for bug reports and other stuff      *
 * --------------------------------------------------------- *
 * This program is free software; you can modify it under    *
 * the terms of the GNU General Public License as published  *
 * by the Free Software Foundation, located in LICENSE       *
 *                                                           *
 *************************************************************/

#ifndef GAME_H
#define GAME_H

#include <iostream>
#include <fstream>
//#include <SDL_gfx/SDL_gfxPrimitives.h>
#include <SDL/SDL.h>
#include <ctime>
#include <vector>
#include <SDL/SDL_ttf.h>
#include <sstream>
#include <ctime>
#include <string>
#include "cl_init.h"
#include "cl_settings.h"
#include "cl_server.h"


class cl_player
{
  public:
    cl_game&              game;
    const unsigned short  keyset;  //>=10 net
    const unsigned short  num;
    cl_bomberman*         bomberman;

    inline cl_player(const cl_player& player);
    inline cl_player(cl_game& game, const unsigned short& keyset, const unsigned short& num);
    inline cl_player& operator= (const cl_player& player);
};

class cl_game
{
  public:
    cl_game(void);
    ~cl_game(void);

    void play(void);
    
  private: 
    typedef  ref_ptr<cl_player>       player_ptr;
    typedef  std::vector<player_ptr>  players_type;

    /* contained objects
     * (they're pointers so that they can be initialized in correct order)
     */
    cl_init*      Inital;
    cl_images*    Images;
    cl_settings*  Settings;
    cl_server*    server;
    /* FIXME: old array limited this to 4 players;
     * maybe do likewise at other place?
     */
    players_type  players;

    /* private methods (non-UI) */
    void init_players(void);

    /* UI specific data */
    unsigned short BB_DISPLAYWIDTH;
    TTF_Font*     display_font;
    SDL_Surface*  text_surface;
    SDL_Surface*  Display;
    SDL_Surface*  BField;

    /* UI specific methods */
    bool parse_key(const SDLKey& key);
    void FillSurface(SDL_Surface* screen, SDL_Surface* texture);
    void DrawDisplay(const int& frlength);
    void DrawBField(void);
    void DrawPlBombermans(void);
    void DrawBombs(void);
    void DrawBomberman(const cl_bomberman& bman);
    void  _M_process_server_events(void);
};

#endif
