/*************************************************************
 *                 bbomberman                                *
 *  (c) 2004 by Bernhard Eder (bb-generation)                *
 *  http://bbgen.net/bomberman for updates                   *
 *  Webmaster@bbgen.net for bug reports and other stuff      *
 * --------------------------------------------------------- *
 * This program is free software; you can modify it under    *
 * the terms of the GNU General Public License as published  *
 * by the Free Software Foundation, located in LICENSE       *
 *                                                           *
 *************************************************************/

#include "cl_proto.h"
/* FIXME: debug */
#include <iostream>

/* only used for internal implementation
 * (therefore not included via cl_proto.h)
 */
#include <sstream>
#include <exception>
#include <stdexcept>


/* set values of string constants defined in cl_proto.h */
const char*  BB_PROTO_STR_DO_BBMOVE    = "DO_BBMOVE";
const char*  BB_PROTO_STR_DO_BBBOMB    = "DO_BBBOMB";

const char*  BB_PROTO_STR_EV_BBMOVE    = "EV_BBMOVE";
const char*  BB_PROTO_STR_EV_BBBOMB    = "EV_BBBOMB";
const char*  BB_PROTO_STR_EV_BBPUP     = "EV_BBPUP";
const char*  BB_PROTO_STR_EV_BBKILL    = "EV_BBKILL";
const char*  BB_PROTO_STR_EV_MAPFIELD  = "EV_MAPFIELD";
const char*  BB_PROTO_STR_EV_MAPCOLUMN = "EV_MAPCOLUMN";
const char*  BB_PROTO_STR_EV_DFCREATE  = "EV_DFCREATE";
const char*  BB_PROTO_STR_EV_DFERASE   = "EV_DFERASE";

const char*  BB_PROTO_STR_DENY_BBMOVE  = "DENY_BBMOVE";
const char*  BB_PROTO_STR_DENY_BBBOMB  = "DENY_BBBOMB";


/*
 * cl_proto_msg implementation
 */

/* ctor: initialization from message type only */
cl_proto_msg::cl_proto_msg(const bb_proto_msg type) :
  _M_type(type)
{ }

/* ctor: initialization from message type & old message */
cl_proto_msg::cl_proto_msg(const bb_proto_msg type, const cl_proto_msg& old_msg) :
  _M_type(type), _M_args(old_msg._M_args)
{ }

/* ctor: initialization from list of public member vars */
cl_proto_msg::cl_proto_msg(const enum bb_proto_msg type, const args_type& args) :
  _M_type(type), _M_args(args)
{ }

/* ctor: initialization from string */
cl_proto_msg::cl_proto_msg(const std::string& msg_str)
{
  /* decode initialization string */
  decode(msg_str);
}

/* check argument index validity */
void cl_proto_msg::_M_check_arg_index(const size_t n) const
{
  if ((n < 0) || (n >= _M_args.size()))
    throw std::runtime_error("cl_proto_msg::arg(): invalid argument index");
}

/* access a single message argument */
std::string& cl_proto_msg::arg(const size_t n)
{
  _M_check_arg_index(n);

  return _M_args[n];
}

/* access a single message argument (read-only) */
const std::string& cl_proto_msg::arg(const size_t n) const
{
  _M_check_arg_index(n);

  return _M_args[n];
}

/* access a single message argument as int (read-only) */
const int cl_proto_msg::arg_to_int(const size_t n) const
{
  _M_check_arg_index(n);

  /* create string stream & initialize it from requested message argument */
  std::stringstream sstream(_M_args[n]);
  int ret;

  /* read int from message argument & return it */
  sstream >> ret;
  return ret;
}

/* access message arguments (read-only) */
cl_proto_msg::args_type cl_proto_msg::args() const
{
  return args_type(_M_args);
}

/* retrieve count of message arguments */
size_t cl_proto_msg::args_count() const
{
  return _M_args.size();
}

/* clear message argument list */
void cl_proto_msg::clear_args()
{
  _M_args.clear();
}

/* decode protocol message from string format */
void cl_proto_msg::decode(const std::string& msg_str_arg)
{
  std::string             msg_str(msg_str_arg);
  std::list<std::string>  tokens;  /* list of tokens in message  */
  std::string             token;   /* current token buffer       */

  /* tokenization: iterate through message string */
  std::string::iterator i = msg_str.begin();
  while (i < msg_str.end())
  {
    /* examine current char */
    switch (*i)
    {
      case '\\':  /* begin of double-char escape sequence? */
        /* discard sequence introduction char */
        i++;

        /* append special char to token buffer */
        switch (*i)
        {
          case '\\':  /* the char itself, without its special meaning */
          case ' ':   /* in tokenization etc.                         */
            token += *i;
            break;
          case 'n':  /* newline */
            token += '\n';
            break;
          default:
            /* ignore unknown escape sequence */
            break;
        }

        break;

      case ' ':  /* end of token? */
        /* ignore leading/trailing whitespace */
        if (!token.empty())
        {
          /* add decoded token to list of tokens in message */
          tokens.push_back(token);
          
          /* clear token buffer */
          token.clear();
        }
        break;

      default:  /* append regular char to token buffer */
        token += *i;
        break;
    }

    /* advance to next char */
    i++;
  }

  /* add last token to token list (has no trailing delimiter) */
  if (!token.empty())
    tokens.push_back(token);
  token.clear();

  /* extract first token: message type */
  if (tokens.empty())
    throw std::runtime_error("cl_proto_msg::decode(): message type missing");
  std::string msg_type = tokens.front();
  tokens.pop_front();


  /* convert message type from string to bb_proto_msg */
  if (msg_type == BB_PROTO_STR_DO_BBMOVE)
    _M_type = BB_PROTO_MSG_DO_BBMOVE;
  else if (msg_type == BB_PROTO_STR_DO_BBBOMB)
    _M_type = BB_PROTO_MSG_DO_BBBOMB;

  else if (msg_type == BB_PROTO_STR_EV_BBMOVE)
    _M_type = BB_PROTO_MSG_EV_BBMOVE;
  else if (msg_type == BB_PROTO_STR_EV_BBBOMB)
    _M_type = BB_PROTO_MSG_EV_BBBOMB;
  else if (msg_type == BB_PROTO_STR_EV_BBPUP)
    _M_type = BB_PROTO_MSG_EV_BBPUP;
  else if (msg_type == BB_PROTO_STR_EV_BBKILL)
    _M_type = BB_PROTO_MSG_EV_BBKILL;

  else if (msg_type == BB_PROTO_STR_EV_MAPFIELD)
    _M_type = BB_PROTO_MSG_EV_MAPFIELD;
  else if (msg_type == BB_PROTO_STR_EV_MAPCOLUMN)
    _M_type = BB_PROTO_MSG_EV_MAPCOLUMN;

  else if (msg_type == BB_PROTO_STR_EV_DFCREATE)
    _M_type = BB_PROTO_MSG_EV_DFCREATE;
  else if (msg_type == BB_PROTO_STR_EV_DFERASE)
    _M_type = BB_PROTO_MSG_EV_DFERASE;

  else if (msg_type == BB_PROTO_STR_DENY_BBMOVE)
    _M_type = BB_PROTO_MSG_EV_BBMOVE;
  else if (msg_type == BB_PROTO_STR_DENY_BBBOMB)
    _M_type = BB_PROTO_MSG_EV_BBBOMB;

  else
    /* invalid message type string */
    throw std::runtime_error("cl_proto_msg::decode(): invalid message type");


  /* assign rest of tokens to argument list member var */
  _M_args.clear();
  std::list<std::string>::iterator j = tokens.begin();
  while (j != tokens.end())
  {
    _M_args.push_back(*j);
    j++;
  }
}

/* encode protocol message into string format (for transport over network) */
std::string cl_proto_msg::encode() const
{
  std::stringstream str_stream;  /* message string buffer */

  /* convert message type from bb_proto_msg to string */
  switch (_M_type)
  {
    case BB_PROTO_MSG_DO_BBMOVE:
      str_stream << BB_PROTO_STR_DO_BBMOVE;
      break;
    case BB_PROTO_MSG_DO_BBBOMB:
      str_stream << BB_PROTO_STR_DO_BBBOMB;
      break;

    case BB_PROTO_MSG_EV_BBMOVE:
      str_stream << BB_PROTO_STR_EV_BBMOVE;
      break;
    case BB_PROTO_MSG_EV_BBBOMB:
      str_stream << BB_PROTO_STR_EV_BBBOMB;
      break;
    case BB_PROTO_MSG_EV_BBPUP:
      str_stream << BB_PROTO_STR_EV_BBPUP;
      break;
    case BB_PROTO_MSG_EV_BBKILL:
      str_stream << BB_PROTO_STR_EV_BBKILL;
      break;

    case BB_PROTO_MSG_EV_MAPFIELD:
      str_stream << BB_PROTO_STR_EV_MAPFIELD;
      break;
    case BB_PROTO_MSG_EV_MAPCOLUMN:
      str_stream << BB_PROTO_STR_EV_MAPCOLUMN;
      break;

    case BB_PROTO_MSG_EV_DFCREATE:
      str_stream << BB_PROTO_STR_EV_DFCREATE;
      break;
    case BB_PROTO_MSG_EV_DFERASE:
      str_stream << BB_PROTO_STR_EV_DFERASE;
      break;

    case BB_PROTO_MSG_DENY_BBMOVE:
      str_stream << BB_PROTO_STR_DENY_BBMOVE;
      break;
    case BB_PROTO_MSG_DENY_BBBOMB:
      str_stream << BB_PROTO_STR_DENY_BBBOMB;
      break;

    default:
      /* invalid message type */
      throw std::runtime_error("cl_proto_msg::encode(): invalid message type");
  }

  /* iterate through message arguments */
  for (args_type::const_iterator i = _M_args.begin(); i < _M_args.end(); i++)
  {
    /* append escaped argument to buffer */
    str_stream << " " << escape(*i);
  }

  /* return stringified message */
  return str_stream.str();
}

/* escape a string, i.e. put backslashs before special chars */
std::string cl_proto_msg::escape(const std::string& str) const
{
  std::string out;

  /* iterate through input string */
  for (std::string::const_iterator i = str.begin(); i < str.end(); i++)
  {
    /* escape special chars as needed */
    switch (*i)
    {
      case ' ':
        out += "\\ ";
        break;
      case '\n':
        out += "\\n";
        break;
      default:
        /* regular char, just append to output string */
        out += *i;
    }
  }

  return out;
}

/* push message argument (of type string) to end of message arguments */
void cl_proto_msg::push_arg(const std::string& arg)
{
  _M_args.push_back(arg);
}

/* push message argument (of type int) to end of message arguments */
void cl_proto_msg::push_arg(const int arg)
{
  /* create string stream & read int into it */
  std::stringstream sstream;
  sstream << arg;

  /* push string stream's contents onto message argument vector */
  _M_args.push_back(sstream.str());
}

/* indicate message type */
bb_proto_msg cl_proto_msg::type() const
{
  return _M_type;
}

/* set new message type */
void cl_proto_msg::type(const bb_proto_msg type)
{
  _M_type = type;
}


/*
 * cl_proto_queue implementation
 */

/* determine if queue is empty */
bool cl_proto_queue::empty() const
{
  return std::list<cl_proto_msg>::empty();
}

/* poll a message from the queue and encode it as string */
std::string cl_proto_queue::poll_encode_msg(void)
{
  return poll_msg().encode();
}

/* poll a message from the queue */
cl_proto_msg cl_proto_queue::poll_msg(void)
{
  cl_proto_msg msg = front();
  pop_front();

  std::cerr << "* cl_proto_queue::poll_msg(" << msg.type();
  std::vector<std::string> args(msg.args());
  for (std::vector<std::string>::iterator arg = args.begin();
       arg < args.end(); arg++)
    std::cerr << ", " << *arg;
  std::cerr << ")" << std::endl;

  return msg;
}

/* decode message string and store message in the queue */
void cl_proto_queue::queue_decode_msg(const std::string& msg_str)
{
  queue_msg(cl_proto_msg(msg_str));
}

/* store message in the queue */
void cl_proto_queue::queue_msg(const cl_proto_msg& msg)
{
  std::cerr << "* cl_proto_queue::queue_msg(" << msg.type();
  std::vector<std::string> args(msg.args());
  for (std::vector<std::string>::iterator arg = args.begin();
       arg < args.end(); arg++)
    std::cerr << ", " << *arg;
  std::cerr << ")" << std::endl;

  push_back(msg);
}

/* compute size of queue */
size_t cl_proto_queue::size() const
{
  return std::list<cl_proto_msg>::size();
}
