/*************************************************************
 *                 bbomberman                                *
 *  (c) 2004 by Bernhard Eder (bb-generation)                *
 *  http://bbgen.net/bomberman for updates                   *
 *  Webmaster@bbgen.net for bug reports and other stuff      *
 * --------------------------------------------------------- *
 * This program is free software; you can modify it under    *
 * the terms of the GNU General Public License as published  *
 * by the Free Software Foundation, located in LICENSE       *
 *                                                           *
 *************************************************************/

#ifndef SERVER_NET_H
#define SERVER_NET_H

/* cl_server_net inherits from cl_server_local,
 * cl_client_net inherits from cl_client
 */
#include "cl_server.h"

/* SDL_net is used for portable networking */
#include <SDL/SDL_net.h>


/* prototype classes defined here (so they can reference each other) */
class cl_client_net;
class cl_server_net;

class cl_client_net : public cl_client
{
  public:
    cl_server_net&  server;

    cl_client_net(cl_server_net& server, TCPsocket socket);
    ~cl_client_net();

    void  detach();
    void  pass_proto_msgs();
    bool  process_net_events();

  private:
    static const int  BB_NET_BUF_LEN = 256;

    TCPsocket  _M_socket;
    bool       _M_connected;
    char       _M_buffer[BB_NET_BUF_LEN];
    char*      _M_buffer_insert_ptr;
};

class cl_server_net : public cl_server_local
{
  friend class cl_client_net;

  public:
    /* default port for BBomberman network connections */
    static const unsigned int  BB_NET_PORT_DEF = 34938;

    cl_server_net(const int map_fields_x, const int map_fields_y, const bomberman_names_type& bomberman_names, const unsigned int port = BB_NET_PORT_DEF);
    ~cl_server_net();

    void  tick();

  protected:
    SDLNet_SocketSet  _M_socket_set;

  private:
    TCPsocket         _M_server_socket;

    void  _M_accept_client();
    void  _M_pass_proto_msgs();
    void  _M_process_net_events(int socket_count);
};

#endif
